.EQU debug=1
; a0: (out) servo stunde
; a1: (out) servo minute
; a2: (out) Lautsprecher (1:Strom an)
; a3: (out) rote LED (1:an)
; a4: (out) gruene LED (1:an)
; a6: (in)  IR Diode
; a7: (in)  Abstandssensor

.ORG 0
                                ; Interrupt Tabelle
        jmp     irq00_reset
        jmp     reset           ; irq01_ext0
        jmp     reset           ; irq02_ext1
        jmp     reset           ; irq03_ext2
        jmp     irq04_timer2cmp
        jmp     irq05_timer2ovf
        jmp     reset           ; irq06_timer1event
        jmp     reset           ; irq07_timer1cmpA
        jmp     reset           ; irq08_timer1cmpB
        jmp     reset           ; irq09_timer1ovf
        jmp     irq0a_timer0cmp
        jmp     irq0b_timer0ovf
        jmp     reset           ; irq0c_spi
        jmp     reset           ; irq0d_V24Rx
        jmp     reset           ; irq0e_V24dat
        jmp     reset           ; irq0f_V24Tx
        jmp     reset           ; irq10_ADC
        jmp     reset           ; irq11_EEPROM
        jmp     reset           ; irq12_analog
        jmp     reset           ; irq13_2wire
        jmp     reset           ; irq14_ProgMem

reset:
irq00_reset:
; ***************** init stack ******************************
        ldi     r16,(ramend >> 8)
        out     sp_hi,r16
        ldi     r16,(ramend & $ff)
        out     sp_lo,r16
; ***********************************************************

; ***************** init ports ******************************
        ldi     r21,$1f         ; a0-a4 output
        out     port_a_dir,r21
; ***********************************************************

; ***************** init V24 ********************************
;    IF debug
        ldi     r16,51
        out     v24_ctr4,r16   ; 9600 baud, 8 MHz CPU

        ldi     r16,0b10000110  ; 1, asyn, no parity, 1 stop, 8 bit, 0
        out     v24_ctr3,r16

        ldi     r16,0b00001000  ; no RX int, no TX int, no empty int, RX disable
        out     v24_ctr2,r16   ; TX enable, 8 data bits, -, 9 bit
;    ENDIF
; ***********************************************************
 
; ***************** init Fernbedienung **********************
        ldi     r16,(buf&$ff)
        sts     buf_ptr,r16
        ldi     r16,(key&$ff)
        sts     key_read,r16
        sts     key_write,r16
        ldi     r16,0
        sts     fb_mode,r16
        dec     r16
        sts     last_key,r16
; ***********************************************************

; ***************** init uhr ********************************
        ldi     r16,0
        sts     ticks,r16
        sts     ticks+1,r16
        sts     ticks+2,r16
; ***********************************************************

; ***************** init a/d wandler ************************
        ldi     r16,0b00100111  ; ext. aref, left adj, pin a7
        out     adc_mux,r16
; ***********************************************************

; ***************** init servos *****************************
        ldi      r31,(servo0 >> 8)
        ldi      r30,(servo0 & $ff)
        sts      servo_ptr,r30
        sts      servo_ptr+1,r31
        
        ldi     r16,8
        ldi     r17,0           ; pos=0
        ldi     r18,95          ; nullpunkt = 95*16us = 1,5 ms
        ldi     r19,1           ; port pin 0
   a10: st      Z+,r17          ; pos
        st      Z+,r18          ; nullpunkt
        st      Z+,r19          ; pin
        add     r19,r19
        dec     r16
        brne    a10
        
        cpi     r18,255
        breq    a20

        ldi     r16,4           ; 4 x
        ldi     r18,255         ; 2 ms pause
        rjmp    a10
   a20: ldi     r16,92          ; Nullpunktkorrektur fuer servo1
        sts     servo1+1,r16
; ***********************************************************

; ***************** init Timer 0 ****************************
;       fuer Uhr, Sound und Fernbedienung
        ldi     r16,0b00001011          ; CTC mode, 8 MHz/64 =125 kHz
        out     timer0_ctr,r16         ; 8 us Takt
        ldi     r16,8                  ; 8*8us=64us
        out     timer0_cmp,r16
        in      r16,timer_ienable
        ori     r16,2                  ; enable Timer0 compare match int
        out     timer_ienable,r16
; ***********************************************************

; ***************** init Timer 2 ****************************
;       fuer Servos
        ldi     r16,0b00001101          ; CTC mode, 8 MHz/128 =62,5 kHz
        out     timer2_ctr,r16         ; 16 us Takt -> 255*16us = 4 ms
        in      r16,timer_ienable
        ori     r16,$80                ; enable Timer0 compare match int
        out     timer_ienable,r16
; ***********************************************************

; ***************** enable Interrupts ***********************
        sei                            ; enable Interrupts
; ***********************************************************

; ***************** Hauptprogramm ***************************
main:   rjmp    main
; ***********************************************************

; ***************** Uhrzeit ausgeben ************************
display:                        ; r25: stunden   r24: minuten
        ret        
; ***********************************************************
        
; ***************** Lesen Fernbedienung *********************
get_key: ; gelesenes Zeichen -> r16
        push    r31
        push    r30
        ldi     r31,(key>>8)
        lds     r30,key_read
   b10: lds     r16,key_write
        cp      r16,r30
        brne    b20
        sleep
        rjmp    b10

   b20: ld      r16,Z+
        mov     r31,r30
        andi    r31,$1f
        brne    b30
        ldi     r30,(key&$ff)
   b30: sts     key_read,r30
        pop     r30
        pop     r31
        ret

tst_key: ; Z=1 fall kein Zeichen vorhanden
        push    r0
        push    r1
        lds     r0,key_read
        lds     r1,key_write
        cp      r1,r0
        pop     r1
        pop     r0
        ret
; ***********************************************************

; ***************** Lesen Abstandssensor ********************
get_abst: ; Abstandssensor -> r16
        ldi     r16,0b11010110     ; adc enable, start conv., no auto trig.,
        out     adc_ctr,r16       ; clear IF, int disable, prescale 64
c10:    sbis    adc_ctr,4         ; conversion ready
        rjmp    c10               ; no
        in      r16,adc_hi
        ret
; ***********************************************************

        
;IF debug
; ***************** Ausgabe V24 *****************************
out_r16:
   d10: sbis    v24_ctr1,5        ; data register empty
        rjmp    d10
        out     v24_dat,r16
        ret

out_hex_r16:
        ret
; ***********************************************************
;ENDIF

; ***************** Timer0 overflow interrupt ***************
irq0b_timer0ovf:
        reti
; ***********************************************************

; ***************** Timer0 compare interrupt ****************
; Aufruf alle 64 us
; Zaehlt Uhr (ticks 32 bit) hoch
; 

irq0a_timer0cmp:
        push    r16
        in      r16,sreg
        push    r16


clock:  ; uncrement ticks, time_m und time_h

ton:    ; Alarmton ausgeben

fernbed:lds     r16,fb_mode
        subi    r16,1
        breq    _0sig           ; fb_mode=1
        brpl    _nosig          ; fb_mode=2
        rjmp    _1sig           ; fb_mode=0        

;---+________/////|/////--\\\\\\|\\\\\\
;   33332222222222211111111110000000000
;   43210987654321098765432109876543210

_nosig: lds     r16,fb_count
        sbis    port_a_pin,6
        rjmp    e10             ; Eingangssignal=0

        dec     r16             ; Wartezeit--
        brpl    e20             ; nicht abgelaufen
        eor     r16,r16         ; nicht negativ
   e20: sts     fb_count,r16
        rjmp    _end1

   e10: or      r16,r16         ; Wartezeit abgelaufen?
        brne    _err1           ; nein
        ldi     r16,1
        sts     fb_mode,r16     ; 0 Signal
        ldi     r16,34
        sts     fb_count,r16    ; max. 34*64us = 2.18 ms "0"
        rjmp    _end1

_0sig:  push    r17
        push    r30
        push    r31
        lds     r16,fb_count
        sbic    port_a_pin,6
        rjmp    e30             ; Eingangssignal=1
        dec     r16             ; max. Laenge ueberschritten
        brmi    _err2           ; ja
        sts     fb_count,r16
        rjmp    _end2

   e30: ldi     r17,0
   e50: ldi     r31,(buf>>8)
        lds     r30,buf_ptr
        cpi     r30,(buf+27)&$ff ; schon 27 0/1 empfangen
        brcc    _err2           ; ja, dann Fehler
        cpi     r16,26          ; weniger als (34-26)*64us = 512 us
        brcc    _err2           ; ja, dann Fehler
        cpi     r16,15          ; 512us < t <1152us 
        brcc    e40             ; ja, dann eine 0/1
        cpi     r16,13          ; 1152us < t y 1408 us
        brcc    _err2           ; ja, dann Fehler
        st      Z+,r17          ; 1408us < t < 2176us : zwei 0/1
   e40: st      Z+,r17
        sts     buf_ptr,r30
        sts     fb_mode,r17     ; 1/0 Signal
        ldi     r16,34
        sts     fb_count,r16    ; max. 34*64us = 2.18 ms "1"
;       rjmp    _end2

_end2:  pop     r31
        pop     r30
        pop     r17
_end1:  pop     r16
        out     sreg,r16
        pop     r16
        reti
        
_err2:  pop     r31
        pop     r30
        pop     r17
_err1:  ldi     r16,2
        sts     fb_mode,r16     ; kein Signal
        ldi     r16,56
        sts     fb_count,r16    ; mindestens 3.5 ms inaktiv       
        ldi     r16,buf&$ff
        sts     buf_ptr,r16     ; buf_ptr zurueck setzen
        rjmp    _end1


_1sig:  push    r17
        push    r30
        push    r31
       
        lds     r16,fb_count
        ldi     r17,1
        sbis    port_a_pin,6
        rjmp    e50             ; Eingangssignal=0
        dec     r16             ; max. Laenge ueberschritten
        sts     fb_count,r16
        brpl    _end2           ; nein, dann weiter warten
        lds     r30,buf_ptr
        cpi     r30,(buf+26)&$ff ; mindestens 26 0/1 empfangen
        brcs    _err2           ; nein, dann Fehler
        lds     r17,buf+3       ; Toggle bit
        add     r17,r17         ; soll MSB werden
        ldi     r31,(buf>>8)
        ldi     r30,(buf+15)&$ff ; Zeiger auf 1. Kommando Bit
        
   e60: ld      r16,Z+          ; naechstes Kommando Bit
        add     r17,r17         ; alter Wert nach links
        or      r17,r16         ; + neues Bit
        inc     r30
        cpi     r30,(buf+26)&$ff ; alle 6 Kommando Bits bearbeitet
        brcs    e60             ; nein, dann weiter

        lds     r16,last_key    ; kein Tastenrepeat
        cp      r17,r16         ; gleiche Taste wie letzte?
        breq    _err2           ; ja, dann ignorieren
        sts     last_key,r17
        andi    r17,$3f         ; nur 6 bit
                
        ldi     r31,key>>8
        lds     r30,key_write   ; Schreibzeiger
        st      Z+,r17          ; empfangenes Kommando abspeichern
        mov     r17,r30
        andi    r17,$1f         ; naechstes zu schr. Byte (0-31)
        brne    e70             ; falls nicht 0, kein wrap around
        ldi     r30,(key&$ff)   ; sonst Pufferanfang eintragen
   e70: lds     r16,key_read
        cpse    r16,r30        ; Lesepointer gleich neuem Schreibpointer
        sts     key_write,r30   ; nein, dann Schreibpointer aktualisieren
        rjmp    _err2

; ***********************************************************


; ***************** Timer2 overflow interrupt ***************
irq05_timer2ovf:
        reti
; ***********************************************************

; ***************** Timer2 compare interrupt ****************
.EQU mask=0b00000011  ; nur  servo0 und servo1

irq04_timer2cmp:
        push    r16
        push    r17
        push    r18
        push    r30
        push    r31
        in      r16,sreg
        push    r16
        
        lds     r30,servo_ptr
        lds     r31,servo_ptr+1

        ld      r16,Z+
        ld      r17,Z+
        add     r17,r16
        ld      r16,Z+
        andi    r16,mask
        in      r18,port_a_dat
        andi    r18,~mask
        or      r16,r18
        out     port_a_dat,r16
        out     timer2_cmp,r17

        cpi     r30,(servo_end & $ff)
        brne    f10
        ldi     r31,(servo0 >> 8)
        ldi     r30,(servo0 & $ff)
   f10: sts     servo_ptr,r30
        sts     servo_ptr+1,r31
                        
           
        pop     r16
        out     sreg,r16
        pop     r31
        pop     r30
        pop     r18
        pop     r17
        pop     r16
        reti

tab_h:  .DB     -48,-40,-32,-24,-16, -8,  0,  8 ,16 ,24, 32,  40, 48, 0
;                 0   1   2   3   4   5   6   7   8   9   10  11  12
tab_m:  .DB      48, 46, 45, 43, 42, 40, 38, 37 ,35 ,34 ; 0x
        .DB      32, 30, 29, 27, 26, 24, 22, 21 ,19 ,18 ; 1x
        .DB      16, 14, 13, 11, 10,  8,  6,  5 , 3 , 2 ; 2x
        .DB       0, -2, -3, -5, -6, -8,-10,-11 ,-13,-14; 3x
        .DB     -16,-18,-19,-21,-22,-24,-26,-27 ,-29,-30; 4x
        .DB     -32,-34,-35,-37,-38,-40,-42,-43 ,-45,-46; 5x

                  
; ***********************************************************

                


;*****************************************************************
;    IO Adressraum
;*****************************************************************

.DSEG
.ORG 0

twbr:           .BYTE   1       ; 00
twsr:           .BYTE   1       ; 01
twar:           .BYTE   1       ; 02
twdr:           .BYTE   1       ; 03
adc_lo:         .BYTE   1       ; 04 ADCL
adc_hi:         .BYTE   1       ; 05 ADCH
adc_ctr:        .BYTE   1       ; 06 ADCSRA
adc_mux:        .BYTE   1       ; 07 ADMUX
acsr:           .BYTE   1       ; 08
v24_ctr4:       .BYTE   1       ; 09 UBRRL
v24_ctr2:       .BYTE   1       ; 0a UCSRB
v24_ctr1:       .BYTE   1       ; 0b UCSRA
v24_dat:        .BYTE   1       ; 0c UDR
spcr:           .BYTE   1       ; 0d
spsr:           .BYTE   1       ; 0e
spdr:           .BYTE   1       ; 0f
port_d_pin:     .BYTE   1       ; 10 PIND
port_d_dir:     .BYTE   1       ; 11 DDRD
port_d_dat:     .BYTE   1       ; 12 PORTD
port_c_pin:     .BYTE   1       ; 13 PINC
port_c_dir:     .BYTE   1       ; 14 DDRC
port_c_dat:     .BYTE   1       ; 15 PORTC
port_b_pin:     .BYTE   1       ; 16 PINB
port_b_dir:     .BYTE   1       ; 17 DDRB
port_b_dat:     .BYTE   1       ; 18 PORTB
port_a_pin:     .BYTE   1       ; 19 PINA
port_a_dir:     .BYTE   1       ; 1a DDRA
port_a_dat:     .BYTE   1       ; 1b PORTA
eecr:           .BYTE   1       ; 1c
eedr:           .BYTE   1       ; 1d
eearl:          .BYTE   1       ; 1e
eearh:          .BYTE   1       ; 1f
v24_ctr3:       .BYTE   1       ; 20 UCSRC
wdtcr:          .BYTE   1       ; 21
assr:           .BYTE   1       ; 22
timer2_cmp:     .BYTE   1       ; 23 OCR2
timer2_dat:     .BYTE   1       ; 24 TCNT2
timer2_ctr:     .BYTE   1       ; 25 TCCR2
icr1l:          .BYTE   1       ; 26
icr1h:          .BYTE   1       ; 27
ocr1bl:         .BYTE   1       ; 28
ocr1bh:         .BYTE   1       ; 29
ocr1al:         .BYTE   1       ; 2a
ocr1ah:         .BYTE   1       ; 2b
tcnt1l:         .BYTE   1       ; 2c
tcnt1h:         .BYTE   1       ; 2d
tccr1b:         .BYTE   1       ; 2e
tccr1a:         .BYTE   1       ; 3f
sfior:          .BYTE   1       ; 30
ocdr:           .BYTE   1       ; 31
timer0_dat:     .BYTE   1       ; 32 TCNT0
timer0_ctr:     .BYTE   1       ; 33 TCCR0
mcucsr:         .BYTE   1       ; 34
mcucr:          .BYTE   1       ; 35
twcr:           .BYTE   1       ; 36
spmcr:          .BYTE   1       ; 37
timer_iflag:    .BYTE   1       ; 38 TIFR
timer_ienable:  .BYTE   1       ; 39 TIMSK
gifr:           .BYTE   1       ; 3a
gicr:           .BYTE   1       ; 3b
timer0_cmp:     .BYTE   1       ; 3c OCR0
sp_lo:          .BYTE   1       ; 3d SPL
sp_hi:          .BYTE   1       ; 3e SPH
sreg:           .BYTE   1       ; 3f

;*****************************************************************
;    SRAM Adressraum
;*****************************************************************

;       .DSEG
;       .ORG 0
;       .BYTE   32      ; register
;       .BYTE   64      ; IO register

.DSEG
.ORG 96

ticks:          .BYTE   4

speak:          .BYTE   1
time_h:         .BYTE   1
time_m:         .BYTE   1
weck_on:        .BYTE   1
weck_h:         .BYTE   1
weck_m:         .BYTE   1

fb_mode:        .BYTE   1
fb_count:       .BYTE   1
buf_ptr:        .BYTE   1
key_read:       .BYTE   1
key_write:      .BYTE   1
last_key:       .BYTE   1

servo_ptr:      .BYTE   2
servo0:         .BYTE   3       ; wert (-128 .. +127); nullpunkt (~175)
servo1:         .BYTE   3
servo2:         .BYTE   3
servo3:         .BYTE   3
servo4:         .BYTE   3
servo5:         .BYTE   3
servo6:         .BYTE   3
servo7:         .BYTE   3
pause0:         .BYTE   3
pause1:         .BYTE   3
pause2:         .BYTE   3
pause3:         .BYTE   3
servo_end:

x:              .BYTE   (((x+31)>>5)<<5)-x
buf:            .BYTE   32
key:            .BYTE   32
.EQU ramend=$85f
